@@ -5,10 +5,40 @@ use utf8;
 use 5.010;
 use Mojo::Base 'WWW::Crawler::Mojo';
 use Mojo::Util qw(md5_sum);
-use WWW::Crawler::Mojo::ScraperUtil qw{html_handlers resolve_href guess_encoding};
-use Mojolicious::Types;
+use Mojo::File;
+use Mojo::Log;
+use WWW::Crawler::Mojo::ScraperUtil qw{resolve_href guess_encoding};
 use Encode;
-our $VERSION = '0.05';
+our $VERSION = '0.06';
+
+my $html_type_regex = qr{^(text|application)/(html|xml|xhtml)};
+my $css_type_regex = qr{^text/css$};
+my $types = {
+    'application/atom+xml'      => 'atom',
+    'application/font-woff'     => 'woff',
+    'application/javascript'    => 'js',
+    'application/json'          => 'json',
+    'application/pdf'           => 'pdf',
+    'application/rss+xml'       => 'rss',
+    'application/x-gzip'        => 'gz',
+    'application/xml'           => 'xml',
+    'application/zip'           => 'zip',
+    'audio/mpeg'                => 'mp3',
+    'audio/ogg'                 => 'ogg',
+    'image/gif'                 => 'gif',
+    'image/jpeg'                => 'jpg',
+    'image/png'                 => 'png',
+    'image/svg+xml'             => 'svg',
+    'image/x-icon'              => 'ico',
+    'text/cache-manifest'       => 'appcache',
+    'text/css'                  => 'css',
+    'text/html'                 => 'html',
+    'text/plain'                => 'txt',
+    'text/xml'                  => 'xml',
+    'video/mp4'                 => 'mp4',
+    'video/ogg'                 => 'ogv',
+    'video/webm'                => 'webm',
+};
 
 has depth => 10;
 has filenames => sub { {} };
@@ -18,14 +48,8 @@ has 'normalize';
 has asset_name => sub { asset_number_generator(6) };
 has _retrys => sub { {} };
 has max_retry => 3;
-has types => sub {
-    my $types;
-    my %cat = %{Mojolicious::Types->new->mapping};
-    while (my ($key, $val) = each %cat) {
-        $types->{$_} = $key for (map { s/\;.*$//; lc $_ } @$val);
-    }
-    return $types;
-};
+has 'log_name';
+has 'log';
 
 sub asset_number_generator {
     my $digit = (shift || 6);
@@ -39,7 +63,7 @@ sub asset_hash_generator {
     my $len = (shift || 6);
     my %uniq;
     return sub {
-        my $md5 = md5_sum(shift);
+        my $md5 = md5_sum(shift->url);
         my $len = $len;
         my $key;
         do { $key = substr($md5, 0, $len++) } while (exists $uniq{$key});
@@ -51,8 +75,16 @@ sub asset_hash_generator {
 sub init {
     my ($self) = @_;
     
+    $self->log(Mojo::Log->new(
+        path => $self->basedir. $self->log_name)) if $self->log_name;
+    
     for (keys %{$self->filenames}) {
         $self->enqueue($_);
+        if (!ref $_) {
+            my $val = $self->filenames->{$_};
+            delete $self->filenames->{$_};
+            $self->filenames->{Mojo::URL->new($_)} = $val;
+        }
     }
     
     $self->on(res => sub {
@@ -62,7 +94,7 @@ sub init {
         
         for my $job2 ($scrape->()) {
             
-            next unless ($self->is_target->($job2, $job2->context));
+            next unless ($self->is_target->($job2));
             next unless ($job2->depth <= $self->depth);
             
             my $url = $job2->url;
@@ -71,15 +103,28 @@ sub init {
                 $job2->url($url = $cb->($url));
             }
             
-            $self->_regist_asset_name($url);
+            next if ($self->filenames->{$url});
+            
+            my $new_id = $self->asset_name->($job2);
+            my $type = $self->ua->head($url)->res->headers->content_type || '';
+            my $ext1 = $types->{ lc(($type =~ /([^;]+)/)[0]) };
+            my $ext2 = ($url->path =~ qr{\.(\w+)$})[0];
+            my $ext = $ext1 || $ext2;
+            $new_id .= ".$ext" if $ext;
+            $self->filenames->{$url} = $new_id;
+            
+            next if $type !~ $html_type_regex && $type !~ $css_type_regex
+                                                && -f $self->basedir. $new_id;
+            
             $self->enqueue($job2);
         }
         
         my $url = $job->url;
         my $type = $res->headers->content_type;
         my $original = $job->original_url;
+        my $save_file = $self->filenames->{$original};
         
-        if ($type && $type =~ qr{text/(html|xml)}) {
+        if ($type && $type =~ $html_type_regex) {
             my $encode = guess_encoding($res) || 'UTF-8';
             my $cont = Mojo::DOM->new(Encode::decode($encode, $res->body));
             my $base = $url;
@@ -88,19 +133,20 @@ sub init {
                 $base = resolve_href($base, $base_tag->attr('href'));
             }
             
-            $self->flatten_html($cont, $base);
+            $self->flatten_html($cont, $base, $save_file);
             $cont = $cont->to_string;
             $self->save($original, $cont, $encode);
-        } elsif ($type && $type =~ qr{text/css}) {
+        } elsif ($type && $type =~ $css_type_regex) {
             my $encode = guess_encoding($res) || 'UTF-8';
-            my $cont = $self->flatten_css($res->body, $url);
+            my $cont = $self->flatten_css($res->body, $url, $save_file);
             $self->save($original, $cont, $encode);
         } else {
             $self->save($original, $res->body);
         }
         
-        say sprintf('created: %s => %s ',
-                                    $self->filenames->{$original}, $original);
+        my $log_line = sprintf('created: %s => %s ', $save_file, $original);
+        say $log_line;
+        $self->log->info($log_line) if $self->log;
     });
     
     $self->on(error => sub {
@@ -117,32 +163,33 @@ sub init {
 }
 
 sub get_href {
-    my ($self, $base, $url) = @_;
+    my ($self, $base, $url, $ref_path) = @_;
     my $fragment = ($url =~ qr{(#.+)})[0] || '';
     my $abs = resolve_href($base, $url);
     if (my $cb = $self->normalize) {
         $abs = $cb->($abs);
     }
     my $file = $self->filenames->{$abs};
-    return './'. $file. $fragment if ($file);
+    my $refdir = Mojo::File->new($ref_path || '')->dirname;
+    return (Mojo::File->new($file)->to_rel($refdir)). $fragment if ($file);
     return $abs. $fragment;
 }
 
 sub flatten_html {
-    my ($self, $dom, $base) = @_;
+    my ($self, $dom, $base, $ref_path) = @_;
     
-    state $handlers = html_handlers();
+    state $handlers = $self->html_handlers();
     $dom->find(join(',', keys %{$handlers}))->each(sub {
         my $dom = shift;
         for ('href', 'ping','src','data') {
-            $dom->{$_} = $self->get_href($base, $dom->{$_}) if ($dom->{$_});
+            $dom->{$_} = $self->get_href($base, $dom->{$_}, $ref_path) if ($dom->{$_});
         }
     });
     
     $dom->find('meta[content]')->each(sub {
         if ($_[0] =~ qr{http\-equiv="?Refresh"?}i && $_[0]->{content}) {
             $_[0]->{content} =~
-                            s{URL=(.+)}{ 'URL='. $self->get_href($base, $1) }e;
+                            s{URL=(.+)}{ 'URL='. $self->get_href($base, $1, $ref_path) }e;
         }
     });
 
@@ -152,47 +199,47 @@ sub flatten_html {
     $dom->find('style')->each(sub {
         my $dom = shift;
         my $cont = $dom->content;
-        $dom->content($self->flatten_css($cont, $base));
+        $dom->content($self->flatten_css($cont, $base, $ref_path));
     });
     
     $dom->find('[style]')->each(sub {
         my $dom = shift;
         my $cont = $dom->{style};
-        $dom->{style} = $self->flatten_css($dom->{style}, $base);
+        $dom->{style} = $self->flatten_css($dom->{style}, $base, $ref_path);
     });
     return $dom
 }
 
 sub flatten_css {
-    my ($self, $cont, $base) = @_;
+    my ($self, $cont, $base, $ref_path) = @_;
     $cont =~ s{url\((.+?)\)}{
         my $url = $1;
         $url =~ s/^(['"])// && $url =~ s/$1$//;
-        'url('. $self->get_href($base, $url). ')';
+        'url('. $self->get_href($base, $url, $ref_path). ')';
     }egi;
     return $cont;
 }
 
 sub save {
     my ($self, $url, $content, $encode) = @_;
-    my $fullpath = $self->basedir. $self->filenames->{$url};
+    my $path =  Mojo::File->new($self->basedir. $self->filenames->{$url});
+    $path->dirname->make_path unless -d $path->dirname;
     $content = Encode::encode($encode, $content) if $encode;
-    open(my $OUT, '>', $fullpath);
-    print $OUT $content;
-    close($OUT);
+    $path->spurt($content);
 }
 
-sub _regist_asset_name {
-    my ($self, $url) = @_;
-    if (!$self->filenames->{$url}) {
-        $self->filenames->{$url} = $self->asset_name->($url);
-        my $ext = ($url->path =~ qr{\.(\w+)$})[0] || do {
-            my $got = $self->ua->head($url)->res->headers->content_type || '';
-            $got =~ s/\;.*$//;
-            $self->types->{lc $got};
-        };
-        $self->filenames->{$url} .= ".$ext" if ($ext);
-    }
+sub say_start {
+    my $self = shift;
+    
+    my $content = <<"EOF";
+----------------------------------------
+Crawling is starting with @{[ $self->queue->next->url ]}
+Max Connection  : @{[ $self->max_conn ]}
+User Agent      : @{[ $self->ua_name ]}
+----------------------------------------
+EOF
+    say $content;
+    $self->log->append($content) if $self->log;
 }
 
 1;
@@ -277,7 +324,7 @@ A directory path for output files.
 Set the condition which indecates whether the job is flatten target or not.
 
     $bot->is_target(sub {
-        my ($job, $context) = @_;
+        my $job = shift;
         ...
         return 1 # or 0
     });
@@ -305,10 +352,6 @@ another option asset_hash_generator, which generates 6 character hash.
 
 Max attempt limit of retry in case the server in inresponsible. Defaults to 3.
 
-=head2 types
-
-MIME types. Defaults to Mojolicious::Types.
-
 =head1 METHODS
 
 =head2 asset_number_generator
